import os
from os import path
import sys

try:
    import simplejson as json
except ImportError:
    import json

from .detector import Detector
from .lang_detect_exception import ErrorCode, LangDetectException
from .utils.lang_profile import LangProfile


class DetectorFactory(object):
    '''
    Language Detector Factory Class.

    This class manages an initialization and constructions of Detector.

    Before using language detection library,
    load profiles with DetectorFactory.load_profile(str)
    and set initialization parameters.

    When the language detection,
    construct Detector instance via DetectorFactory.create().
    See also Detector's sample code.
    '''
    seed = None

    def __init__(self):
        self.word_lang_prob_map = {}
        self.langlist = []

    def load_profile(self, profile_directory):
        list_files = os.listdir(profile_directory)
        if not list_files:
            raise LangDetectException(ErrorCode.NeedLoadProfileError, 'Not found profile: ' + profile_directory)

        langsize, index = len(list_files), 0
        for filename in list_files:
            if filename.startswith('.'):
                continue
            filename = path.join(profile_directory, filename)
            if not path.isfile(filename):
                continue

            f = None
            try:
                if sys.version_info[0] < 3:
                    f = open(filename, 'r')
                else:
                    f = open(filename, 'r', encoding='utf-8')
                json_data = json.load(f)
                profile = LangProfile(**json_data)
                self.add_profile(profile, index, langsize)
                index += 1
            except IOError:
                raise LangDetectException(ErrorCode.FileLoadError, 'Cannot open "%s"' % filename)
            except Exception:
                raise LangDetectException(ErrorCode.FormatError, 'Profile format error in "%s"' % filename)
            finally:
                if f:
                    f.close()

    def load_json_profile(self, json_profiles):
        langsize, index = len(json_profiles), 0
        if langsize < 2:
            raise LangDetectException(ErrorCode.NeedLoadProfileError, 'Need more than 2 profiles.')

        for json_profile in json_profiles:
            try:
                json_data = json.loads(json_profile)
                profile = LangProfile(**json_data)
                self.add_profile(profile, index, langsize)
                index += 1
            except Exception:
                raise LangDetectException(ErrorCode.FormatError, 'Profile format error.')

    def add_profile(self, profile, index, langsize):
        lang = profile.name
        if lang in self.langlist:
            raise LangDetectException(ErrorCode.DuplicateLangError, 'Duplicate the same language profile.')
        self.langlist.append(lang)

        for word in profile.freq:
            if word not in self.word_lang_prob_map:
                self.word_lang_prob_map[word] = [0.0] * langsize
            length = len(word)
            if 1 <= length <= 3:
                prob = 1.0 * profile.freq.get(word) / profile.n_words[length - 1]
                self.word_lang_prob_map[word][index] = prob

    def clear(self):
        self.langlist = []
        self.word_lang_prob_map = {}

    def create(self, alpha=None):
        '''Construct Detector instance with smoothing parameter.'''
        detector = self._create_detector()
        if alpha is not None:
            detector.set_alpha(alpha)
        return detector

    def _create_detector(self):
        if not self.langlist:
            raise LangDetectException(ErrorCode.NeedLoadProfileError, 'Need to load profiles.')
        return Detector(self)

    def set_seed(self, seed):
        self.seed = seed

    def get_lang_list(self):
        return list(self.langlist)


PROFILES_DIRECTORY = path.join(path.dirname(__file__), 'profiles')
_factory = None

def init_factory():
    global _factory
    if _factory is None:
        _factory = DetectorFactory()
        _factory.load_profile(PROFILES_DIRECTORY)

def detect(text):
    init_factory()
    detector = _factory.create()
    detector.append(text)
    return detector.detect()


def detect_langs(text):
    init_factory()
    detector = _factory.create()
    detector.append(text)
    return detector.get_probabilities()



_error_codes = {
    'NoTextError': 0,
    'FormatError': 1,
    'FileLoadError': 2,
    'DuplicateLangError': 3,
    'NeedLoadProfileError': 4,
    'CantDetectError': 5,
    'CantOpenTrainData': 6,
    'TrainDataFormatError': 7,
    'InitParamError': 8,
}

ErrorCode = type('ErrorCode', (), _error_codes)


class LangDetectException(Exception):
    def __init__(self, code, message):
        super(LangDetectException, self).__init__(message)
        self.code = code

    def get_code(self):
        return self.code



import unittest

import six

from langdetect_zh.detector_factory import DetectorFactory
from langdetect_zh.utils.lang_profile import LangProfile


class DetectorTest(unittest.TestCase):
    TRAINING_EN = 'a a a b b c c d e'
    TRAINING_FR = 'a b b c c c d d d'
    TRAINING_JA = six.u('\u3042 \u3042 \u3042 \u3044 \u3046 \u3048 \u3048')
    JSON_LANG1 = '{"freq":{"A":3,"B":6,"C":3,"AB":2,"BC":1,"ABC":2,"BBC":1,"CBA":1},"n_words":[12,3,4],"name":"lang1"}'
    JSON_LANG2 = '{"freq":{"A":6,"B":3,"C":3,"AA":3,"AB":2,"ABC":1,"ABA":1,"CAA":1},"n_words":[12,5,3],"name":"lang2"}'

    def setUp(self):
        self.factory = DetectorFactory()

        profile_en = LangProfile('en')
        for w in self.TRAINING_EN.split():
            profile_en.add(w)
        self.factory.add_profile(profile_en, 0, 3)

        profile_fr = LangProfile('fr')
        for w in self.TRAINING_FR.split():
            profile_fr.add(w)
        self.factory.add_profile(profile_fr, 1, 3)

        profile_ja = LangProfile('ja')
        for w in self.TRAINING_JA.split():
            profile_ja.add(w)
        self.factory.add_profile(profile_ja, 2, 3)

    def test_detector1(self):
        detect = self.factory.create()
        detect.append('a')
        self.assertEqual(detect.detect(), 'en')

    def test_detector2(self):
        detect = self.factory.create()
        detect.append('b d')
        self.assertEqual(detect.detect(), 'fr')

    def test_detector3(self):
        detect = self.factory.create()
        detect.append('d e')
        self.assertEqual(detect.detect(), 'en')

    def test_detector4(self):
        detect = self.factory.create()
        detect.append(six.u('\u3042\u3042\u3042\u3042a'))
        self.assertEqual(detect.detect(), 'ja')

    def test_lang_list(self):
        langlist = self.factory.get_lang_list()
        self.assertEqual(len(langlist), 3)
        self.assertEqual(langlist[0], 'en')
        self.assertEqual(langlist[1], 'fr')
        self.assertEqual(langlist[2], 'ja')

    def test_factory_from_json_string(self):
        self.factory.clear()
        profiles = [self.JSON_LANG1, self.JSON_LANG2]
        self.factory.load_json_profile(profiles)
        langlist = self.factory.get_lang_list()
        self.assertEqual(len(langlist), 2)
        self.assertEqual(langlist[0], 'lang1')
        self.assertEqual(langlist[1], 'lang2')



class Language(object):
    '''
    Language is to store the detected language.
    Detector.get_probabilities() returns a list of Languages.
    '''

    def __init__(self, lang, prob):
        self.lang = lang
        self.prob = prob

    def __repr__(self):
        if self.lang is None:
            return ''
        return '%s:%s' % (self.lang, self.prob)

    def __lt__(self, other):
        return self.prob < other.prob



import random
import re

import six
from six.moves import zip, xrange

from .lang_detect_exception import ErrorCode, LangDetectException
from .language import Language
from .utils.ngram import NGram
from .utils.unicode_block import unicode_block


class Detector(object):
    '''
    Detector class is to detect language from specified text.
    Its instance is able to be constructed via the factory class DetectorFactory.

    After appending a target text to the Detector instance with .append(string),
    the detector provides the language detection results for target text via .detect() or .get_probabilities().

    .detect() method returns a single language name which has the highest probability.
    .get_probabilities() methods returns a list of multiple languages and their probabilities.

    The detector has some parameters for language detection.
    See set_alpha(double), .set_max_text_length(int) .set_prior_map(dict).

    Example:

        from langdetect.detector_factory import DetectorFactory
        factory = DetectorFactory()
        factory.load_profile('/path/to/profile/directory')

        def detect(text):
            detector = factory.create()
            detector.append(text)
            return detector.detect()

        def detect_langs(text):
            detector = factory.create()
            detector.append(text)
            return detector.get_probabilities()
    '''

    ALPHA_DEFAULT = 0.5
    ALPHA_WIDTH = 0.05

    ITERATION_LIMIT = 1000
    PROB_THRESHOLD = 0.1
    CONV_THRESHOLD = 0.99999
    BASE_FREQ = 10000
    UNKNOWN_LANG = 'unknown'

    URL_RE = re.compile(r'https?://[-_.?&~;+=/#0-9A-Za-z]{1,2076}')
    MAIL_RE = re.compile(r'[-_.0-9A-Za-z]{1,64}@[-_0-9A-Za-z]{1,255}[-_.0-9A-Za-z]{1,255}')

    def __init__(self, factory):
        self.word_lang_prob_map = factory.word_lang_prob_map
        self.langlist = factory.langlist
        self.seed = factory.seed
        self.random = random.Random()
        self.text = ''
        self.langprob = None

        self.alpha = self.ALPHA_DEFAULT
        self.n_trial = 7
        self.max_text_length = 10000
        self.prior_map = None
        self.verbose = False

    def set_verbose(self):
        self.verbose = True

    def set_alpha(self, alpha):
        self.alpha = alpha

    def set_prior_map(self, prior_map):
        '''Set prior information about language probabilities.'''
        self.prior_map = [0.0] * len(self.langlist)
        sump = 0.0
        for i in xrange(len(self.prior_map)):
            lang = self.langlist[i]
            if lang in prior_map:
                p = prior_map[lang]
                if p < 0:
                    raise LangDetectException(ErrorCode.InitParamError, 'Prior probability must be non-negative.')
                self.prior_map[i] = p
                sump += p
        if sump <= 0.0:
            raise LangDetectException(ErrorCode.InitParamError, 'More one of prior probability must be non-zero.')
        for i in xrange(len(self.prior_map)):
            self.prior_map[i] /= sump

    def set_max_text_length(self, max_text_length):
        '''Specify max size of target text to use for language detection.
        The default value is 10000(10KB).
        '''
        self.max_text_length = max_text_length

    def append(self, text):
        '''Append the target text for language detection.
        If the total size of target text exceeds the limit size specified by
        Detector.set_max_text_length(int), the rest is cut down.
        '''
        text = self.URL_RE.sub(' ', text)
        text = self.MAIL_RE.sub(' ', text)
        text = NGram.normalize_vi(text)
        pre = 0
        for i in xrange(min(len(text), self.max_text_length)):
            ch = text[i]
            if ch != ' ' or pre != ' ':
                self.text += ch
            pre = ch

    def cleaning_text(self):
        '''Cleaning text to detect
        (eliminate URL, e-mail address and Latin sentence if it is not written in Latin alphabet).
        '''
        latin_count, non_latin_count = 0, 0
        for ch in self.text:
            if 'A' <= ch <= 'z':
                latin_count += 1
            elif ch >= six.u('\u0300') and unicode_block(ch) != 'Latin Extended Additional':
                non_latin_count += 1

        if latin_count * 2 < non_latin_count:
            text_without_latin = ''
            for ch in self.text:
                if ch < 'A' or 'z' < ch:
                    text_without_latin += ch
            self.text = text_without_latin

    def detect(self):
        '''Detect language of the target text and return the language name
        which has the highest probability.
        '''
        probabilities = self.get_probabilities()
        if probabilities:
            return probabilities[0].lang
        return self.UNKNOWN_LANG

    def get_probabilities(self):
        if self.langprob is None:
            self._detect_block()
        return self._sort_probability(self.langprob)

    def _detect_block(self):
        self.cleaning_text()
        ngrams = self._extract_ngrams()
        if not ngrams:
            raise LangDetectException(ErrorCode.CantDetectError, 'No features in text.')

        self.langprob = [0.0] * len(self.langlist)

        self.random.seed(self.seed)
        for t in xrange(self.n_trial):
            prob = self._init_probability()
            alpha = self.alpha + self.random.gauss(0.0, 1.0) * self.ALPHA_WIDTH

            i = 0
            while True:
                self._update_lang_prob(prob, self.random.choice(ngrams), alpha)
                if i % 5 == 0:
                    if self._normalize_prob(prob) > self.CONV_THRESHOLD or i >= self.ITERATION_LIMIT:
                        break
                    if self.verbose:
                        six.print_('>', self._sort_probability(prob))
                i += 1
            for j in xrange(len(self.langprob)):
                self.langprob[j] += prob[j] / self.n_trial
            if self.verbose:
                six.print_('==>', self._sort_probability(prob))

    def _init_probability(self):
        '''Initialize the map of language probabilities.
        If there is the specified prior map, use it as initial map.
        '''
        if self.prior_map is not None:
            return list(self.prior_map)
        else:
            return [1.0 / len(self.langlist)] * len(self.langlist)

    def _extract_ngrams(self):
        '''Extract n-grams from target text.'''
        RANGE = list(xrange(1, NGram.N_GRAM + 1))

        result = []
        ngram = NGram()
        for ch in self.text:
            ngram.add_char(ch)
            if ngram.capitalword:
                continue
            for n in RANGE:
                # optimized w = ngram.get(n)
                if len(ngram.grams) < n:
                    break
                w = ngram.grams[-n:]
                if w and w != ' ' and w in self.word_lang_prob_map:
                    result.append(w)
        return result

    def _update_lang_prob(self, prob, word, alpha):
        '''Update language probabilities with N-gram string(N=1,2,3).'''
        if word is None or word not in self.word_lang_prob_map:
            return False

        lang_prob_map = self.word_lang_prob_map[word]
        if self.verbose:
            six.print_('%s(%s): %s' % (word, self._unicode_encode(word), self._word_prob_to_string(lang_prob_map)))

        weight = alpha / self.BASE_FREQ
        for i in xrange(len(prob)):
            prob[i] *= weight + lang_prob_map[i]
        return True

    def _word_prob_to_string(self, prob):
        result = ''
        for j in xrange(len(prob)):
            p = prob[j]
            if p >= 0.00001:
                result += ' %s:%.5f' % (self.langlist[j], p)
        return result

    def _normalize_prob(self, prob):
        '''Normalize probabilities and check convergence by the maximun probability.
        '''
        maxp, sump = 0.0, sum(prob)
        for i in xrange(len(prob)):
            p = prob[i] / sump
            if maxp < p:
                maxp = p
            prob[i] = p
        return maxp

    def _sort_probability(self, prob):
        result = [Language(lang, p) for (lang, p) in zip(self.langlist, prob) if p > self.PROB_THRESHOLD]
        result.sort(reverse=True)
        return result

    def _unicode_encode(self, word):
        buf = ''
        for ch in word:
            if ch >= six.u('\u0080'):
                st = hex(0x10000 + ord(ch))[2:]
                while len(st) < 4:
                    st = '0' + st
                buf += r'\u' + st[1:5]
            else:
                buf += ch
        return buf



import unittest

from langdetect_zh.language import Language


class LanguageTest(unittest.TestCase):
    def test_language(self):
        lang = Language(None, 0)
        self.assertIsNone(lang.lang)
        self.assertEqual(lang.prob, 0.0, 0.0001)
        self.assertEqual(str(lang), '')

        lang2 = Language('en', 1.0)
        self.assertEqual(lang2.lang, 'en')
        self.assertEqual(lang2.prob, 1.0, 0.0001)
        self.assertEqual(str(lang2), 'en:1.0')

    def test_cmp(self):
        lang1 = Language('a', 0.1)
        lang2 = Language('b', 0.5)

        self.assertTrue(lang1 < lang2)
        self.assertFalse(lang1 == lang2)
        self.assertFalse(lang1 > lang1)



from collections import defaultdict
import re

import six
from six.moves import xrange

from .ngram import NGram


class LangProfile(object):
    MINIMUM_FREQ = 2
    LESS_FREQ_RATIO = 100000

    ROMAN_CHAR_RE = re.compile(r'^[A-Za-z]$')
    ROMAN_SUBSTR_RE = re.compile(r'.*[A-Za-z].*')

    def __init__(self, name=None, freq=None, n_words=None):
        self.freq = defaultdict(int)
        if freq is not None:
            self.freq.update(freq)

        if n_words is None:
            n_words = [0] * NGram.N_GRAM

        self.name = name
        self.n_words = n_words

    def add(self, gram):
        '''Add n-gram to profile.'''
        if self.name is None or gram is None:  # Illegal
            return
        length = len(gram)
        if length < 1 or length > NGram.N_GRAM:  # Illegal
            return
        self.n_words[length - 1] += 1
        self.freq[gram] += 1

    def omit_less_freq(self):
        '''Eliminate below less frequency n-grams and noise Latin alphabets.'''
        if self.name is None:  # Illegal
            return
        threshold = max(self.n_words[0] // self.LESS_FREQ_RATIO, self.MINIMUM_FREQ)

        roman = 0
        for key, count in list(six.iteritems(self.freq)):
            if count <= threshold:
                self.n_words[len(key)-1] -= count
                del self.freq[key]
            elif self.ROMAN_CHAR_RE.match(key):
                roman += count

        # roman check
        if roman < self.n_words[0] // 3:
            for key, count in list(six.iteritems(self.freq)):
                if self.ROMAN_SUBSTR_RE.match(key):
                    self.n_words[len(key)-1] -= count
                    del self.freq[key]

    def update(self, text):
        '''Update the language profile with (fragmented) text.
        Extract n-grams from text and add their frequency into the profile.
        '''
        if text is None:
            return
        text = NGram.normalize_vi(text)
        gram = NGram()
        for ch in text:
            gram.add_char(ch)
            for n in xrange(1, NGram.N_GRAM+1):
                self.add(gram.get(n))



from os import path


class Messages(object):
    MESSAGES_FILENAME = path.join(path.dirname(__file__), 'messages.properties')

    def __init__(self):
        self.messages = {}
        with open(self.MESSAGES_FILENAME, 'r') as f:
            for line in f:
                key, _, value = line.strip().partition('=')
                self.messages[key] = value.encode().decode('unicode_escape')

    def get_string(self, key):
        return self.messages.get(key, '!%s!' % key)


_messages = None
def get_string(key):
    global _messages
    if _messages is None:
        _messages = Messages()
    return _messages.get_string(key)



UNICODE_BASIC_LATIN = 1
UNICODE_LATIN_1_SUPPLEMENT = 2
UNICODE_LATIN_EXTENDED_A = 3
UNICODE_LATIN_EXTENDED_B = 4
UNICODE_IPA_EXTENSIONS = 5
UNICODE_SPACING_MODIFIER_LETTERS = 6
UNICODE_COMBINING_DIACRITICAL_MARKS = 7
UNICODE_GREEK_AND_COPTIC = 8
UNICODE_CYRILLIC = 9
UNICODE_CYRILLIC_SUPPLEMENT = 10
UNICODE_ARMENIAN = 11
UNICODE_HEBREW = 12
UNICODE_ARABIC = 13
UNICODE_SYRIAC = 14
UNICODE_ARABIC_SUPPLEMENT = 15
UNICODE_THAANA = 16
UNICODE_NKO = 17
UNICODE_SAMARITAN = 18
UNICODE_MANDAIC = 19
UNICODE_ARABIC_EXTENDED_A = 20
UNICODE_DEVANAGARI = 21
UNICODE_BENGALI = 22
UNICODE_GURMUKHI = 23
UNICODE_GUJARATI = 24
UNICODE_ORIYA = 25
UNICODE_TAMIL = 26
UNICODE_TELUGU = 27
UNICODE_KANNADA = 28
UNICODE_MALAYALAM = 29
UNICODE_SINHALA = 30
UNICODE_THAI = 31
UNICODE_LAO = 32
UNICODE_TIBETAN = 33
UNICODE_MYANMAR = 34
UNICODE_GEORGIAN = 35
UNICODE_HANGUL_JAMO = 36
UNICODE_ETHIOPIC = 37
UNICODE_ETHIOPIC_SUPPLEMENT = 38
UNICODE_CHEROKEE = 39
UNICODE_UNIFIED_CANADIAN_ABORIGINAL_SYLLABICS = 40
UNICODE_OGHAM = 41
UNICODE_RUNIC = 42
UNICODE_TAGALOG = 43
UNICODE_HANUNOO = 44
UNICODE_BUHID = 45
UNICODE_TAGBANWA = 46
UNICODE_KHMER = 47
UNICODE_MONGOLIAN = 48
UNICODE_UNIFIED_CANADIAN_ABORIGINAL_SYLLABICS_EXTENDED = 49
UNICODE_LIMBU = 50
UNICODE_TAI_LE = 51
UNICODE_NEW_TAI_LUE = 52
UNICODE_KHMER_SYMBOLS = 53
UNICODE_BUGINESE = 54
UNICODE_TAI_THAM = 55
UNICODE_BALINESE = 56
UNICODE_SUNDANESE = 57
UNICODE_BATAK = 58
UNICODE_LEPCHA = 59
UNICODE_OL_CHIKI = 60
UNICODE_SUNDANESE_SUPPLEMENT = 61
UNICODE_VEDIC_EXTENSIONS = 62
UNICODE_PHONETIC_EXTENSIONS = 63
UNICODE_PHONETIC_EXTENSIONS_SUPPLEMENT = 64
UNICODE_COMBINING_DIACRITICAL_MARKS_SUPPLEMENT = 65
UNICODE_LATIN_EXTENDED_ADDITIONAL = 66
UNICODE_GREEK_EXTENDED = 67
UNICODE_GENERAL_PUNCTUATION = 68
UNICODE_SUPERSCRIPTS_AND_SUBSCRIPTS = 69
UNICODE_CURRENCY_SYMBOLS = 70
UNICODE_COMBINING_DIACRITICAL_MARKS_FOR_SYMBOLS = 71
UNICODE_LETTERLIKE_SYMBOLS = 72
UNICODE_NUMBER_FORMS = 73
UNICODE_ARROWS = 74
UNICODE_MATHEMATICAL_OPERATORS = 75
UNICODE_MISCELLANEOUS_TECHNICAL = 76
UNICODE_CONTROL_PICTURES = 77
UNICODE_OPTICAL_CHARACTER_RECOGNITION = 78
UNICODE_ENCLOSED_ALPHANUMERICS = 79
UNICODE_BOX_DRAWING = 80
UNICODE_BLOCK_ELEMENTS = 81
UNICODE_GEOMETRIC_SHAPES = 82
UNICODE_MISCELLANEOUS_SYMBOLS = 83
UNICODE_DINGBATS = 84
UNICODE_MISCELLANEOUS_MATHEMATICAL_SYMBOLS_A = 85
UNICODE_SUPPLEMENTAL_ARROWS_A = 86
UNICODE_BRAILLE_PATTERNS = 87
UNICODE_SUPPLEMENTAL_ARROWS_B = 88
UNICODE_MISCELLANEOUS_MATHEMATICAL_SYMBOLS_B = 89
UNICODE_SUPPLEMENTAL_MATHEMATICAL_OPERATORS = 90
UNICODE_MISCELLANEOUS_SYMBOLS_AND_ARROWS = 91
UNICODE_GLAGOLITIC = 92
UNICODE_LATIN_EXTENDED_C = 93
UNICODE_COPTIC = 94
UNICODE_GEORGIAN_SUPPLEMENT = 95
UNICODE_TIFINAGH = 96
UNICODE_ETHIOPIC_EXTENDED = 97
UNICODE_CYRILLIC_EXTENDED_A = 98
UNICODE_SUPPLEMENTAL_PUNCTUATION = 99
UNICODE_CJK_RADICALS_SUPPLEMENT = 100
UNICODE_KANGXI_RADICALS = 101
UNICODE_IDEOGRAPHIC_DESCRIPTION_CHARACTERS = 102
UNICODE_CJK_SYMBOLS_AND_PUNCTUATION = 103
UNICODE_HIRAGANA = 104
UNICODE_KATAKANA = 105
UNICODE_BOPOMOFO = 106
UNICODE_HANGUL_COMPATIBILITY_JAMO = 107
UNICODE_KANBUN = 108
UNICODE_BOPOMOFO_EXTENDED = 109
UNICODE_CJK_STROKES = 110
UNICODE_KATAKANA_PHONETIC_EXTENSIONS = 111
UNICODE_ENCLOSED_CJK_LETTERS_AND_MONTHS = 112
UNICODE_CJK_COMPATIBILITY = 113
UNICODE_CJK_UNIFIED_IDEOGRAPHS_EXTENSION_A = 114
UNICODE_YIJING_HEXAGRAM_SYMBOLS = 115
UNICODE_CJK_UNIFIED_IDEOGRAPHS = 116
UNICODE_YI_SYLLABLES = 117
UNICODE_YI_RADICALS = 118
UNICODE_LISU = 119
UNICODE_VAI = 120
UNICODE_CYRILLIC_EXTENDED_B = 121
UNICODE_BAMUM = 122
UNICODE_MODIFIER_TONE_LETTERS = 123
UNICODE_LATIN_EXTENDED_D = 124
UNICODE_SYLOTI_NAGRI = 125
UNICODE_COMMON_INDIC_NUMBER_FORMS = 126
UNICODE_PHAGS_PA = 127
UNICODE_SAURASHTRA = 128
UNICODE_DEVANAGARI_EXTENDED = 129
UNICODE_KAYAH_LI = 130
UNICODE_REJANG = 131
UNICODE_HANGUL_JAMO_EXTENDED_A = 132
UNICODE_JAVANESE = 133
UNICODE_CHAM = 134
UNICODE_MYANMAR_EXTENDED_A = 135
UNICODE_TAI_VIET = 136
UNICODE_MEETEI_MAYEK_EXTENSIONS = 137
UNICODE_ETHIOPIC_EXTENDED_A = 138
UNICODE_MEETEI_MAYEK = 139
UNICODE_HANGUL_SYLLABLES = 140
UNICODE_HANGUL_JAMO_EXTENDED_B = 141
UNICODE_HIGH_SURROGATES = 142
UNICODE_HIGH_PRIVATE_USE_SURROGATES = 143
UNICODE_LOW_SURROGATES = 144
UNICODE_PRIVATE_USE_AREA = 145
UNICODE_CJK_COMPATIBILITY_IDEOGRAPHS = 146
UNICODE_ALPHABETIC_PRESENTATION_FORMS = 147
UNICODE_ARABIC_PRESENTATION_FORMS_A = 148
UNICODE_VARIATION_SELECTORS = 149
UNICODE_VERTICAL_FORMS = 150
UNICODE_COMBINING_HALF_MARKS = 151
UNICODE_CJK_COMPATIBILITY_FORMS = 152
UNICODE_SMALL_FORM_VARIANTS = 153
UNICODE_ARABIC_PRESENTATION_FORMS_B = 154
UNICODE_HALFWIDTH_AND_FULLWIDTH_FORMS = 155
UNICODE_SPECIALS = 156
UNICODE_LINEAR_B_SYLLABARY = 157
UNICODE_LINEAR_B_IDEOGRAMS = 158
UNICODE_AEGEAN_NUMBERS = 159
UNICODE_ANCIENT_GREEK_NUMBERS = 160
UNICODE_ANCIENT_SYMBOLS = 161
UNICODE_PHAISTOS_DISC = 162
UNICODE_LYCIAN = 163
UNICODE_CARIAN = 164
UNICODE_OLD_ITALIC = 165
UNICODE_GOTHIC = 166
UNICODE_UGARITIC = 167
UNICODE_OLD_PERSIAN = 168
UNICODE_DESERET = 169
UNICODE_SHAVIAN = 170
UNICODE_OSMANYA = 171
UNICODE_CYPRIOT_SYLLABARY = 172
UNICODE_IMPERIAL_ARAMAIC = 173
UNICODE_PHOENICIAN = 174
UNICODE_LYDIAN = 175
UNICODE_MEROITIC_HIEROGLYPHS = 176
UNICODE_MEROITIC_CURSIVE = 177
UNICODE_KHAROSHTHI = 178
UNICODE_OLD_SOUTH_ARABIAN = 179
UNICODE_AVESTAN = 180
UNICODE_INSCRIPTIONAL_PARTHIAN = 181
UNICODE_INSCRIPTIONAL_PAHLAVI = 182
UNICODE_OLD_TURKIC = 183
UNICODE_RUMI_NUMERAL_SYMBOLS = 184
UNICODE_BRAHMI = 185
UNICODE_KAITHI = 186
UNICODE_SORA_SOMPENG = 187
UNICODE_CHAKMA = 188
UNICODE_SHARADA = 189
UNICODE_TAKRI = 190
UNICODE_CUNEIFORM = 191
UNICODE_CUNEIFORM_NUMBERS_AND_PUNCTUATION = 192
UNICODE_EGYPTIAN_HIEROGLYPHS = 193
UNICODE_BAMUM_SUPPLEMENT = 194
UNICODE_MIAO = 195
UNICODE_KANA_SUPPLEMENT = 196
UNICODE_BYZANTINE_MUSICAL_SYMBOLS = 197
UNICODE_MUSICAL_SYMBOLS = 198
UNICODE_ANCIENT_GREEK_MUSICAL_NOTATION = 199
UNICODE_TAI_XUAN_JING_SYMBOLS = 200
UNICODE_COUNTING_ROD_NUMERALS = 201
UNICODE_MATHEMATICAL_ALPHANUMERIC_SYMBOLS = 202
UNICODE_ARABIC_MATHEMATICAL_ALPHABETIC_SYMBOLS = 203
UNICODE_MAHJONG_TILES = 204
UNICODE_DOMINO_TILES = 205
UNICODE_PLAYING_CARDS = 206
UNICODE_ENCLOSED_ALPHANUMERIC_SUPPLEMENT = 207
UNICODE_ENCLOSED_IDEOGRAPHIC_SUPPLEMENT = 208
UNICODE_MISCELLANEOUS_SYMBOLS_AND_PICTOGRAPHS = 209
UNICODE_EMOTICONS = 210
UNICODE_TRANSPORT_AND_MAP_SYMBOLS = 211
UNICODE_ALCHEMICAL_SYMBOLS = 212
UNICODE_CJK_UNIFIED_IDEOGRAPHS_EXTENSION_B = 213
UNICODE_CJK_UNIFIED_IDEOGRAPHS_EXTENSION_C = 214
UNICODE_CJK_UNIFIED_IDEOGRAPHS_EXTENSION_D = 215
UNICODE_CJK_COMPATIBILITY_IDEOGRAPHS_SUPPLEMENT = 216
UNICODE_TAGS = 217
UNICODE_VARIATION_SELECTORS_SUPPLEMENT = 218
UNICODE_SUPPLEMENTARY_PRIVATE_USE_AREA_A = 219
UNICODE_SUPPLEMENTARY_PRIVATE_USE_AREA_B = 220


_unicode_blocks = [
    (UNICODE_BASIC_LATIN, 0x0000, 0x007F),
    (UNICODE_LATIN_1_SUPPLEMENT, 0x0080, 0x00FF),
    (UNICODE_LATIN_EXTENDED_A, 0x0100, 0x017F),
    (UNICODE_LATIN_EXTENDED_B, 0x0180, 0x024F),
    (UNICODE_IPA_EXTENSIONS, 0x0250, 0x02AF),
    (UNICODE_SPACING_MODIFIER_LETTERS, 0x02B0, 0x02FF),
    (UNICODE_COMBINING_DIACRITICAL_MARKS, 0x0300, 0x036F),
    (UNICODE_GREEK_AND_COPTIC, 0x0370, 0x03FF),
    (UNICODE_CYRILLIC, 0x0400, 0x04FF),
    (UNICODE_CYRILLIC_SUPPLEMENT, 0x0500, 0x052F),
    (UNICODE_ARMENIAN, 0x0530, 0x058F),
    (UNICODE_HEBREW, 0x0590, 0x05FF),
    (UNICODE_ARABIC, 0x0600, 0x06FF),
    (UNICODE_SYRIAC, 0x0700, 0x074F),
    (UNICODE_ARABIC_SUPPLEMENT, 0x0750, 0x077F),
    (UNICODE_THAANA, 0x0780, 0x07BF),
    (UNICODE_NKO, 0x07C0, 0x07FF),
    (UNICODE_SAMARITAN, 0x0800, 0x083F),
    (UNICODE_MANDAIC, 0x0840, 0x085F),
    (UNICODE_ARABIC_EXTENDED_A, 0x08A0, 0x08FF),
    (UNICODE_DEVANAGARI, 0x0900, 0x097F),
    (UNICODE_BENGALI, 0x0980, 0x09FF),
    (UNICODE_GURMUKHI, 0x0A00, 0x0A7F),
    (UNICODE_GUJARATI, 0x0A80, 0x0AFF),
    (UNICODE_ORIYA, 0x0B00, 0x0B7F),
    (UNICODE_TAMIL, 0x0B80, 0x0BFF),
    (UNICODE_TELUGU, 0x0C00, 0x0C7F),
    (UNICODE_KANNADA, 0x0C80, 0x0CFF),
    (UNICODE_MALAYALAM, 0x0D00, 0x0D7F),
    (UNICODE_SINHALA, 0x0D80, 0x0DFF),
    (UNICODE_THAI, 0x0E00, 0x0E7F),
    (UNICODE_LAO, 0x0E80, 0x0EFF),
    (UNICODE_TIBETAN, 0x0F00, 0x0FFF),
    (UNICODE_MYANMAR, 0x1000, 0x109F),
    (UNICODE_GEORGIAN, 0x10A0, 0x10FF),
    (UNICODE_HANGUL_JAMO, 0x1100, 0x11FF),
    (UNICODE_ETHIOPIC, 0x1200, 0x137F),
    (UNICODE_ETHIOPIC_SUPPLEMENT, 0x1380, 0x139F),
    (UNICODE_CHEROKEE, 0x13A0, 0x13FF),
    (UNICODE_UNIFIED_CANADIAN_ABORIGINAL_SYLLABICS, 0x1400, 0x167F),
    (UNICODE_OGHAM, 0x1680, 0x169F),
    (UNICODE_RUNIC, 0x16A0, 0x16FF),
    (UNICODE_TAGALOG, 0x1700, 0x171F),
    (UNICODE_HANUNOO, 0x1720, 0x173F),
    (UNICODE_BUHID, 0x1740, 0x175F),
    (UNICODE_TAGBANWA, 0x1760, 0x177F),
    (UNICODE_KHMER, 0x1780, 0x17FF),
    (UNICODE_MONGOLIAN, 0x1800, 0x18AF),
    (UNICODE_UNIFIED_CANADIAN_ABORIGINAL_SYLLABICS_EXTENDED, 0x18B0, 0x18FF),
    (UNICODE_LIMBU, 0x1900, 0x194F),
    (UNICODE_TAI_LE, 0x1950, 0x197F),
    (UNICODE_NEW_TAI_LUE, 0x1980, 0x19DF),
    (UNICODE_KHMER_SYMBOLS, 0x19E0, 0x19FF),
    (UNICODE_BUGINESE, 0x1A00, 0x1A1F),
    (UNICODE_TAI_THAM, 0x1A20, 0x1AAF),
    (UNICODE_BALINESE, 0x1B00, 0x1B7F),
    (UNICODE_SUNDANESE, 0x1B80, 0x1BBF),
    (UNICODE_BATAK, 0x1BC0, 0x1BFF),
    (UNICODE_LEPCHA, 0x1C00, 0x1C4F),
    (UNICODE_OL_CHIKI, 0x1C50, 0x1C7F),
    (UNICODE_SUNDANESE_SUPPLEMENT, 0x1CC0, 0x1CCF),
    (UNICODE_VEDIC_EXTENSIONS, 0x1CD0, 0x1CFF),
    (UNICODE_PHONETIC_EXTENSIONS, 0x1D00, 0x1D7F),
    (UNICODE_PHONETIC_EXTENSIONS_SUPPLEMENT, 0x1D80, 0x1DBF),
    (UNICODE_COMBINING_DIACRITICAL_MARKS_SUPPLEMENT, 0x1DC0, 0x1DFF),
    (UNICODE_LATIN_EXTENDED_ADDITIONAL, 0x1E00, 0x1EFF),
    (UNICODE_GREEK_EXTENDED, 0x1F00, 0x1FFF),
    (UNICODE_GENERAL_PUNCTUATION, 0x2000, 0x206F),
    (UNICODE_SUPERSCRIPTS_AND_SUBSCRIPTS, 0x2070, 0x209F),
    (UNICODE_CURRENCY_SYMBOLS, 0x20A0, 0x20CF),
    (UNICODE_COMBINING_DIACRITICAL_MARKS_FOR_SYMBOLS, 0x20D0, 0x20FF),
    (UNICODE_LETTERLIKE_SYMBOLS, 0x2100, 0x214F),
    (UNICODE_NUMBER_FORMS, 0x2150, 0x218F),
    (UNICODE_ARROWS, 0x2190, 0x21FF),
    (UNICODE_MATHEMATICAL_OPERATORS, 0x2200, 0x22FF),
    (UNICODE_MISCELLANEOUS_TECHNICAL, 0x2300, 0x23FF),
    (UNICODE_CONTROL_PICTURES, 0x2400, 0x243F),
    (UNICODE_OPTICAL_CHARACTER_RECOGNITION, 0x2440, 0x245F),
    (UNICODE_ENCLOSED_ALPHANUMERICS, 0x2460, 0x24FF),
    (UNICODE_BOX_DRAWING, 0x2500, 0x257F),
    (UNICODE_BLOCK_ELEMENTS, 0x2580, 0x259F),
    (UNICODE_GEOMETRIC_SHAPES, 0x25A0, 0x25FF),
    (UNICODE_MISCELLANEOUS_SYMBOLS, 0x2600, 0x26FF),
    (UNICODE_DINGBATS, 0x2700, 0x27BF),
    (UNICODE_MISCELLANEOUS_MATHEMATICAL_SYMBOLS_A, 0x27C0, 0x27EF),
    (UNICODE_SUPPLEMENTAL_ARROWS_A, 0x27F0, 0x27FF),
    (UNICODE_BRAILLE_PATTERNS, 0x2800, 0x28FF),
    (UNICODE_SUPPLEMENTAL_ARROWS_B, 0x2900, 0x297F),
    (UNICODE_MISCELLANEOUS_MATHEMATICAL_SYMBOLS_B, 0x2980, 0x29FF),
    (UNICODE_SUPPLEMENTAL_MATHEMATICAL_OPERATORS, 0x2A00, 0x2AFF),
    (UNICODE_MISCELLANEOUS_SYMBOLS_AND_ARROWS, 0x2B00, 0x2BFF),
    (UNICODE_GLAGOLITIC, 0x2C00, 0x2C5F),
    (UNICODE_LATIN_EXTENDED_C, 0x2C60, 0x2C7F),
    (UNICODE_COPTIC, 0x2C80, 0x2CFF),
    (UNICODE_GEORGIAN_SUPPLEMENT, 0x2D00, 0x2D2F),
    (UNICODE_TIFINAGH, 0x2D30, 0x2D7F),
    (UNICODE_ETHIOPIC_EXTENDED, 0x2D80, 0x2DDF),
    (UNICODE_CYRILLIC_EXTENDED_A, 0x2DE0, 0x2DFF),
    (UNICODE_SUPPLEMENTAL_PUNCTUATION, 0x2E00, 0x2E7F),
    (UNICODE_CJK_RADICALS_SUPPLEMENT, 0x2E80, 0x2EFF),
    (UNICODE_KANGXI_RADICALS, 0x2F00, 0x2FDF),
    (UNICODE_IDEOGRAPHIC_DESCRIPTION_CHARACTERS, 0x2FF0, 0x2FFF),
    (UNICODE_CJK_SYMBOLS_AND_PUNCTUATION, 0x3000, 0x303F),
    (UNICODE_HIRAGANA, 0x3040, 0x309F),
    (UNICODE_KATAKANA, 0x30A0, 0x30FF),
    (UNICODE_BOPOMOFO, 0x3100, 0x312F),
    (UNICODE_HANGUL_COMPATIBILITY_JAMO, 0x3130, 0x318F),
    (UNICODE_KANBUN, 0x3190, 0x319F),
    (UNICODE_BOPOMOFO_EXTENDED, 0x31A0, 0x31BF),
    (UNICODE_CJK_STROKES, 0x31C0, 0x31EF),
    (UNICODE_KATAKANA_PHONETIC_EXTENSIONS, 0x31F0, 0x31FF),
    (UNICODE_ENCLOSED_CJK_LETTERS_AND_MONTHS, 0x3200, 0x32FF),
    (UNICODE_CJK_COMPATIBILITY, 0x3300, 0x33FF),
    (UNICODE_CJK_UNIFIED_IDEOGRAPHS_EXTENSION_A, 0x3400, 0x4DBF),
    (UNICODE_YIJING_HEXAGRAM_SYMBOLS, 0x4DC0, 0x4DFF),
    (UNICODE_CJK_UNIFIED_IDEOGRAPHS, 0x4E00, 0x9FFF),
    (UNICODE_YI_SYLLABLES, 0xA000, 0xA48F),
    (UNICODE_YI_RADICALS, 0xA490, 0xA4CF),
    (UNICODE_LISU, 0xA4D0, 0xA4FF),
    (UNICODE_VAI, 0xA500, 0xA63F),
    (UNICODE_CYRILLIC_EXTENDED_B, 0xA640, 0xA69F),
    (UNICODE_BAMUM, 0xA6A0, 0xA6FF),
    (UNICODE_MODIFIER_TONE_LETTERS, 0xA700, 0xA71F),
    (UNICODE_LATIN_EXTENDED_D, 0xA720, 0xA7FF),
    (UNICODE_SYLOTI_NAGRI, 0xA800, 0xA82F),
    (UNICODE_COMMON_INDIC_NUMBER_FORMS, 0xA830, 0xA83F),
    (UNICODE_PHAGS_PA, 0xA840, 0xA87F),
    (UNICODE_SAURASHTRA, 0xA880, 0xA8DF),
    (UNICODE_DEVANAGARI_EXTENDED, 0xA8E0, 0xA8FF),
    (UNICODE_KAYAH_LI, 0xA900, 0xA92F),
    (UNICODE_REJANG, 0xA930, 0xA95F),
    (UNICODE_HANGUL_JAMO_EXTENDED_A, 0xA960, 0xA97F),
    (UNICODE_JAVANESE, 0xA980, 0xA9DF),
    (UNICODE_CHAM, 0xAA00, 0xAA5F),
    (UNICODE_MYANMAR_EXTENDED_A, 0xAA60, 0xAA7F),
    (UNICODE_TAI_VIET, 0xAA80, 0xAADF),
    (UNICODE_MEETEI_MAYEK_EXTENSIONS, 0xAAE0, 0xAAFF),
    (UNICODE_ETHIOPIC_EXTENDED_A, 0xAB00, 0xAB2F),
    (UNICODE_MEETEI_MAYEK, 0xABC0, 0xABFF),
    (UNICODE_HANGUL_SYLLABLES, 0xAC00, 0xD7AF),
    (UNICODE_HANGUL_JAMO_EXTENDED_B, 0xD7B0, 0xD7FF),
    (UNICODE_HIGH_SURROGATES, 0xD800, 0xDB7F),
    (UNICODE_HIGH_PRIVATE_USE_SURROGATES, 0xDB80, 0xDBFF),
    (UNICODE_LOW_SURROGATES, 0xDC00, 0xDFFF),
    (UNICODE_PRIVATE_USE_AREA, 0xE000, 0xF8FF),
    (UNICODE_CJK_COMPATIBILITY_IDEOGRAPHS, 0xF900, 0xFAFF),
    (UNICODE_ALPHABETIC_PRESENTATION_FORMS, 0xFB00, 0xFB4F),
    (UNICODE_ARABIC_PRESENTATION_FORMS_A, 0xFB50, 0xFDFF),
    (UNICODE_VARIATION_SELECTORS, 0xFE00, 0xFE0F),
    (UNICODE_VERTICAL_FORMS, 0xFE10, 0xFE1F),
    (UNICODE_COMBINING_HALF_MARKS, 0xFE20, 0xFE2F),
    (UNICODE_CJK_COMPATIBILITY_FORMS, 0xFE30, 0xFE4F),
    (UNICODE_SMALL_FORM_VARIANTS, 0xFE50, 0xFE6F),
    (UNICODE_ARABIC_PRESENTATION_FORMS_B, 0xFE70, 0xFEFF),
    (UNICODE_HALFWIDTH_AND_FULLWIDTH_FORMS, 0xFF00, 0xFFEF),
    (UNICODE_SPECIALS, 0xFFF0, 0xFFFF),
    (UNICODE_LINEAR_B_SYLLABARY, 0x10000, 0x1007F),
    (UNICODE_LINEAR_B_IDEOGRAMS, 0x10080, 0x100FF),
    (UNICODE_AEGEAN_NUMBERS, 0x10100, 0x1013F),
    (UNICODE_ANCIENT_GREEK_NUMBERS, 0x10140, 0x1018F),
    (UNICODE_ANCIENT_SYMBOLS, 0x10190, 0x101CF),
    (UNICODE_PHAISTOS_DISC, 0x101D0, 0x101FF),
    (UNICODE_LYCIAN, 0x10280, 0x1029F),
    (UNICODE_CARIAN, 0x102A0, 0x102DF),
    (UNICODE_OLD_ITALIC, 0x10300, 0x1032F),
    (UNICODE_GOTHIC, 0x10330, 0x1034F),
    (UNICODE_UGARITIC, 0x10380, 0x1039F),
    (UNICODE_OLD_PERSIAN, 0x103A0, 0x103DF),
    (UNICODE_DESERET, 0x10400, 0x1044F),
    (UNICODE_SHAVIAN, 0x10450, 0x1047F),
    (UNICODE_OSMANYA, 0x10480, 0x104AF),
    (UNICODE_CYPRIOT_SYLLABARY, 0x10800, 0x1083F),
    (UNICODE_IMPERIAL_ARAMAIC, 0x10840, 0x1085F),
    (UNICODE_PHOENICIAN, 0x10900, 0x1091F),
    (UNICODE_LYDIAN, 0x10920, 0x1093F),
    (UNICODE_MEROITIC_HIEROGLYPHS, 0x10980, 0x1099F),
    (UNICODE_MEROITIC_CURSIVE, 0x109A0, 0x109FF),
    (UNICODE_KHAROSHTHI, 0x10A00, 0x10A5F),
    (UNICODE_OLD_SOUTH_ARABIAN, 0x10A60, 0x10A7F),
    (UNICODE_AVESTAN, 0x10B00, 0x10B3F),
    (UNICODE_INSCRIPTIONAL_PARTHIAN, 0x10B40, 0x10B5F),
    (UNICODE_INSCRIPTIONAL_PAHLAVI, 0x10B60, 0x10B7F),
    (UNICODE_OLD_TURKIC, 0x10C00, 0x10C4F),
    (UNICODE_RUMI_NUMERAL_SYMBOLS, 0x10E60, 0x10E7F),
    (UNICODE_BRAHMI, 0x11000, 0x1107F),
    (UNICODE_KAITHI, 0x11080, 0x110CF),
    (UNICODE_SORA_SOMPENG, 0x110D0, 0x110FF),
    (UNICODE_CHAKMA, 0x11100, 0x1114F),
    (UNICODE_SHARADA, 0x11180, 0x111DF),
    (UNICODE_TAKRI, 0x11680, 0x116CF),
    (UNICODE_CUNEIFORM, 0x12000, 0x123FF),
    (UNICODE_CUNEIFORM_NUMBERS_AND_PUNCTUATION, 0x12400, 0x1247F),
    (UNICODE_EGYPTIAN_HIEROGLYPHS, 0x13000, 0x1342F),
    (UNICODE_BAMUM_SUPPLEMENT, 0x16800, 0x16A3F),
    (UNICODE_MIAO, 0x16F00, 0x16F9F),
    (UNICODE_KANA_SUPPLEMENT, 0x1B000, 0x1B0FF),
    (UNICODE_BYZANTINE_MUSICAL_SYMBOLS, 0x1D000, 0x1D0FF),
    (UNICODE_MUSICAL_SYMBOLS, 0x1D100, 0x1D1FF),
    (UNICODE_ANCIENT_GREEK_MUSICAL_NOTATION, 0x1D200, 0x1D24F),
    (UNICODE_TAI_XUAN_JING_SYMBOLS, 0x1D300, 0x1D35F),
    (UNICODE_COUNTING_ROD_NUMERALS, 0x1D360, 0x1D37F),
    (UNICODE_MATHEMATICAL_ALPHANUMERIC_SYMBOLS, 0x1D400, 0x1D7FF),
    (UNICODE_ARABIC_MATHEMATICAL_ALPHABETIC_SYMBOLS, 0x1EE00, 0x1EEFF),
    (UNICODE_MAHJONG_TILES, 0x1F000, 0x1F02F),
    (UNICODE_DOMINO_TILES, 0x1F030, 0x1F09F),
    (UNICODE_PLAYING_CARDS, 0x1F0A0, 0x1F0FF),
    (UNICODE_ENCLOSED_ALPHANUMERIC_SUPPLEMENT, 0x1F100, 0x1F1FF),
    (UNICODE_ENCLOSED_IDEOGRAPHIC_SUPPLEMENT, 0x1F200, 0x1F2FF),
    (UNICODE_MISCELLANEOUS_SYMBOLS_AND_PICTOGRAPHS, 0x1F300, 0x1F5FF),
    (UNICODE_EMOTICONS, 0x1F600, 0x1F64F),
    (UNICODE_TRANSPORT_AND_MAP_SYMBOLS, 0x1F680, 0x1F6FF),
    (UNICODE_ALCHEMICAL_SYMBOLS, 0x1F700, 0x1F77F),
    (UNICODE_CJK_UNIFIED_IDEOGRAPHS_EXTENSION_B, 0x20000, 0x2A6DF),
    (UNICODE_CJK_UNIFIED_IDEOGRAPHS_EXTENSION_C, 0x2A700, 0x2B73F),
    (UNICODE_CJK_UNIFIED_IDEOGRAPHS_EXTENSION_D, 0x2B740, 0x2B81F),
    (UNICODE_CJK_COMPATIBILITY_IDEOGRAPHS_SUPPLEMENT, 0x2F800, 0x2FA1F),
    (UNICODE_TAGS, 0xE0000, 0xE007F),
    (UNICODE_VARIATION_SELECTORS_SUPPLEMENT, 0xE0100, 0xE01EF),
    (UNICODE_SUPPLEMENTARY_PRIVATE_USE_AREA_A, 0xF0000, 0xFFFFF),
    (UNICODE_SUPPLEMENTARY_PRIVATE_USE_AREA_B, 0x100000, 0x10FFFF),
]

NUM_BLOCKS = len(_unicode_blocks)


def unicode_block(ch):
    '''Return the Unicode block name for ch, or None if ch has no block.'''
    cp = ord(ch)
    # special case basic latin
    if cp <= 0x7F:
        return UNICODE_BASIC_LATIN
    # binary search for the correct block
    be, en = 0, NUM_BLOCKS - 1
    while be <= en:
        mid = (be+en) >> 1
        name, start, end = _unicode_blocks[mid]
        if start <= cp <= end:
            return name
        if cp < start:
            en = mid-1
        else:
            be = mid+1



import re

import six

from . import messages
from .unicode_block import (
    unicode_block,
    UNICODE_BASIC_LATIN,
    UNICODE_LATIN_1_SUPPLEMENT,
    UNICODE_LATIN_EXTENDED_B,
    UNICODE_GENERAL_PUNCTUATION,
    UNICODE_ARABIC,
    UNICODE_LATIN_EXTENDED_ADDITIONAL,
    UNICODE_HIRAGANA,
    UNICODE_KATAKANA,
    UNICODE_BOPOMOFO,
    UNICODE_BOPOMOFO_EXTENDED,
    UNICODE_CJK_UNIFIED_IDEOGRAPHS,
    UNICODE_HANGUL_SYLLABLES,
)


class NGram(object):
    LATIN1_EXCLUDED = messages.get_string('NGram.LATIN1_EXCLUDE')
    N_GRAM = 3

    def __init__(self):
        self.grams = ' '
        self.capitalword = False

    def add_char(self, ch):
        '''Append a character into ngram buffer.'''
        ch = self.normalize(ch)
        last_char = self.grams[-1]
        if last_char == ' ':
            self.grams = ' '
            self.capitalword = False
            if ch == ' ':
                return
        elif len(self.grams) >= self.N_GRAM:
            self.grams = self.grams[1:]
        self.grams += ch

        if ch.isupper():
            if last_char.isupper():
                self.capitalword = True
        else:
            self.capitalword = False

    def get(self, n):
        '''Get n-gram.'''
        if self.capitalword:
            return
        if n < 1 or n > self.N_GRAM or len(self.grams) < n:
            return
        if n == 1:
            ch = self.grams[-1]
            if ch == ' ':
                return
            return ch
        else:
            return self.grams[-n:]

    @classmethod
    def normalize(cls, ch):
        block = unicode_block(ch)
        if block == UNICODE_BASIC_LATIN:
            if ch < 'A' or ('Z' < ch < 'a') or 'z' < ch:
                ch = ' '
        elif block == UNICODE_LATIN_1_SUPPLEMENT:
            if cls.LATIN1_EXCLUDED.find(ch) >= 0:
                ch = ' '
        elif block == UNICODE_LATIN_EXTENDED_B:
            # normalization for Romanian
            if ch == six.u('\u0219'):  # Small S with comma below => with cedilla
                ch = six.u('\u015f')
            if ch == six.u('\u021b'):  # Small T with comma below => with cedilla
                ch = six.u('\u0163')
        elif block == UNICODE_GENERAL_PUNCTUATION:
            ch = ' '
        elif block == UNICODE_ARABIC:
            if ch == six.u('\u06cc'):
                ch = six.u('\u064a')  # Farsi yeh => Arabic yeh
        elif block == UNICODE_LATIN_EXTENDED_ADDITIONAL:
            if ch >= six.u('\u1ea0'):
                ch = six.u('\u1ec3')
        elif block == UNICODE_HIRAGANA:
            ch = six.u('\u3042')
        elif block == UNICODE_KATAKANA:
            ch = six.u('\u30a2')
        elif block in (UNICODE_BOPOMOFO, UNICODE_BOPOMOFO_EXTENDED):
            ch = six.u('\u3105')
        elif block == UNICODE_CJK_UNIFIED_IDEOGRAPHS:
            ch = cls.CJK_MAP.get(ch, ch)
        elif block == UNICODE_HANGUL_SYLLABLES:
            ch = six.u('\uac00')
        return ch

    @classmethod
    def normalize_vi(cls, text):
        '''Normalizer for Vietnamese.
        Normalize Alphabet + Diacritical Mark(U+03xx) into U+1Exx.
        '''
        def repl(m):
            alphabet = cls.TO_NORMALIZE_VI_CHARS.find(m.group(1))
            dmark = cls.DMARK_CLASS.find(m.group(2))  # Diacritical Mark
            return cls.NORMALIZED_VI_CHARS[dmark][alphabet]
        return cls.ALPHABET_WITH_DMARK.sub(repl, text)

    NORMALIZED_VI_CHARS = [
        messages.get_string('NORMALIZED_VI_CHARS_0300'),
        messages.get_string('NORMALIZED_VI_CHARS_0301'),
        messages.get_string('NORMALIZED_VI_CHARS_0303'),
        messages.get_string('NORMALIZED_VI_CHARS_0309'),
        messages.get_string('NORMALIZED_VI_CHARS_0323')]
    TO_NORMALIZE_VI_CHARS = messages.get_string('TO_NORMALIZE_VI_CHARS')
    DMARK_CLASS = messages.get_string('DMARK_CLASS')
    ALPHABET_WITH_DMARK = re.compile(
        '([' + TO_NORMALIZE_VI_CHARS + '])([' + DMARK_CLASS + '])',
        re.UNICODE)

    # CJK Kanji Normalization Mapping
    CJK_CLASS = [
        messages.get_string('NGram.KANJI_1_0'),
        messages.get_string('NGram.KANJI_1_2'),
        messages.get_string('NGram.KANJI_1_4'),
        messages.get_string('NGram.KANJI_1_8'),
        messages.get_string('NGram.KANJI_1_11'),
        messages.get_string('NGram.KANJI_1_12'),
        messages.get_string('NGram.KANJI_1_13'),
        messages.get_string('NGram.KANJI_1_14'),
        messages.get_string('NGram.KANJI_1_16'),
        messages.get_string('NGram.KANJI_1_18'),
        messages.get_string('NGram.KANJI_1_22'),
        messages.get_string('NGram.KANJI_1_27'),
        messages.get_string('NGram.KANJI_1_29'),
        messages.get_string('NGram.KANJI_1_31'),
        messages.get_string('NGram.KANJI_1_35'),
        messages.get_string('NGram.KANJI_2_0'),
        messages.get_string('NGram.KANJI_2_1'),
        messages.get_string('NGram.KANJI_2_4'),
        messages.get_string('NGram.KANJI_2_9'),
        messages.get_string('NGram.KANJI_2_10'),
        messages.get_string('NGram.KANJI_2_11'),
        messages.get_string('NGram.KANJI_2_12'),
        messages.get_string('NGram.KANJI_2_13'),
        messages.get_string('NGram.KANJI_2_15'),
        messages.get_string('NGram.KANJI_2_16'),
        messages.get_string('NGram.KANJI_2_18'),
        messages.get_string('NGram.KANJI_2_21'),
        messages.get_string('NGram.KANJI_2_22'),
        messages.get_string('NGram.KANJI_2_23'),
        messages.get_string('NGram.KANJI_2_28'),
        messages.get_string('NGram.KANJI_2_29'),
        messages.get_string('NGram.KANJI_2_30'),
        messages.get_string('NGram.KANJI_2_31'),
        messages.get_string('NGram.KANJI_2_32'),
        messages.get_string('NGram.KANJI_2_35'),
        messages.get_string('NGram.KANJI_2_36'),
        messages.get_string('NGram.KANJI_2_37'),
        messages.get_string('NGram.KANJI_2_38'),
        messages.get_string('NGram.KANJI_3_1'),
        messages.get_string('NGram.KANJI_3_2'),
        messages.get_string('NGram.KANJI_3_3'),
        messages.get_string('NGram.KANJI_3_4'),
        messages.get_string('NGram.KANJI_3_5'),
        messages.get_string('NGram.KANJI_3_8'),
        messages.get_string('NGram.KANJI_3_9'),
        messages.get_string('NGram.KANJI_3_11'),
        messages.get_string('NGram.KANJI_3_12'),
        messages.get_string('NGram.KANJI_3_13'),
        messages.get_string('NGram.KANJI_3_15'),
        messages.get_string('NGram.KANJI_3_16'),
        messages.get_string('NGram.KANJI_3_18'),
        messages.get_string('NGram.KANJI_3_19'),
        messages.get_string('NGram.KANJI_3_22'),
        messages.get_string('NGram.KANJI_3_23'),
        messages.get_string('NGram.KANJI_3_27'),
        messages.get_string('NGram.KANJI_3_29'),
        messages.get_string('NGram.KANJI_3_30'),
        messages.get_string('NGram.KANJI_3_31'),
        messages.get_string('NGram.KANJI_3_32'),
        messages.get_string('NGram.KANJI_3_35'),
        messages.get_string('NGram.KANJI_3_36'),
        messages.get_string('NGram.KANJI_3_37'),
        messages.get_string('NGram.KANJI_3_38'),
        messages.get_string('NGram.KANJI_4_0'),
        messages.get_string('NGram.KANJI_4_9'),
        messages.get_string('NGram.KANJI_4_10'),
        messages.get_string('NGram.KANJI_4_16'),
        messages.get_string('NGram.KANJI_4_17'),
        messages.get_string('NGram.KANJI_4_18'),
        messages.get_string('NGram.KANJI_4_22'),
        messages.get_string('NGram.KANJI_4_24'),
        messages.get_string('NGram.KANJI_4_28'),
        messages.get_string('NGram.KANJI_4_34'),
        messages.get_string('NGram.KANJI_4_39'),
        messages.get_string('NGram.KANJI_5_10'),
        messages.get_string('NGram.KANJI_5_11'),
        messages.get_string('NGram.KANJI_5_12'),
        messages.get_string('NGram.KANJI_5_13'),
        messages.get_string('NGram.KANJI_5_14'),
        messages.get_string('NGram.KANJI_5_18'),
        messages.get_string('NGram.KANJI_5_26'),
        messages.get_string('NGram.KANJI_5_29'),
        messages.get_string('NGram.KANJI_5_34'),
        messages.get_string('NGram.KANJI_5_39'),
        messages.get_string('NGram.KANJI_6_0'),
        messages.get_string('NGram.KANJI_6_3'),
        messages.get_string('NGram.KANJI_6_9'),
        messages.get_string('NGram.KANJI_6_10'),
        messages.get_string('NGram.KANJI_6_11'),
        messages.get_string('NGram.KANJI_6_12'),
        messages.get_string('NGram.KANJI_6_16'),
        messages.get_string('NGram.KANJI_6_18'),
        messages.get_string('NGram.KANJI_6_20'),
        messages.get_string('NGram.KANJI_6_21'),
        messages.get_string('NGram.KANJI_6_22'),
        messages.get_string('NGram.KANJI_6_23'),
        messages.get_string('NGram.KANJI_6_25'),
        messages.get_string('NGram.KANJI_6_28'),
        messages.get_string('NGram.KANJI_6_29'),
        messages.get_string('NGram.KANJI_6_30'),
        messages.get_string('NGram.KANJI_6_32'),
        messages.get_string('NGram.KANJI_6_34'),
        messages.get_string('NGram.KANJI_6_35'),
        messages.get_string('NGram.KANJI_6_37'),
        messages.get_string('NGram.KANJI_6_39'),
        messages.get_string('NGram.KANJI_7_0'),
        messages.get_string('NGram.KANJI_7_3'),
        messages.get_string('NGram.KANJI_7_6'),
        messages.get_string('NGram.KANJI_7_7'),
        messages.get_string('NGram.KANJI_7_9'),
        messages.get_string('NGram.KANJI_7_11'),
        messages.get_string('NGram.KANJI_7_12'),
        messages.get_string('NGram.KANJI_7_13'),
        messages.get_string('NGram.KANJI_7_16'),
        messages.get_string('NGram.KANJI_7_18'),
        messages.get_string('NGram.KANJI_7_19'),
        messages.get_string('NGram.KANJI_7_20'),
        messages.get_string('NGram.KANJI_7_21'),
        messages.get_string('NGram.KANJI_7_23'),
        messages.get_string('NGram.KANJI_7_25'),
        messages.get_string('NGram.KANJI_7_28'),
        messages.get_string('NGram.KANJI_7_29'),
        messages.get_string('NGram.KANJI_7_32'),
        messages.get_string('NGram.KANJI_7_33'),
        messages.get_string('NGram.KANJI_7_35'),
        messages.get_string('NGram.KANJI_7_37')]

    CJK_MAP = {}

    @classmethod
    def _init_cjk_map(cls):
        for cjk_list in cls.CJK_CLASS:
            representative = cjk_list[0]
            for ch in cjk_list:
                cls.CJK_MAP[ch] = representative

NGram._init_cjk_map()



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

with open('README.md') as f:
    readme = f.read()

setup(
    name='langdetect_zh',
    version='1.0.4',
    description='Google\'s langdetect modified for Chinese texts',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='huyc',
    author_email='huyc@mail.ecust.edu.cn',
    keywords='language detection chinese',
    packages=['langdetect_zh', 'langdetect_zh.utils', 'langdetect_zh.tests'],
    include_package_data=True,
    install_requires=['six'],
    license='MIT',
    license_file='LICENSE',  # 关键修复点
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',  # 修正分类器
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ]
)